@@ -15,7 +15,7 @@ import           Data.Bool (bool)
 import qualified Data.ByteString as BS
 import           Data.Coerce
 import           Data.Foldable (traverse_)
-import           Data.GI.Base (unsafeCastTo)
+import qualified Data.GI.Base.ManagedPtr as ManagedPtr
 import           Data.GI.Base.GError
 import           Data.Int
 import           Data.List
@@ -24,7 +24,6 @@ import           Data.Maybe
 import           Data.Ord
 import           Data.Ratio
 import qualified Data.Text as T
-import qualified GI.DbusmenuGtk3.Objects.Menu as DM
 import qualified GI.GLib as GLib
 import           GI.GLib.Structs.Bytes
 import qualified GI.Gdk as Gdk
@@ -37,12 +36,14 @@ import qualified GI.Gtk as Gtk
 import           GI.Gtk.Flags
 import           GI.Gtk.Objects.IconTheme
 import           Graphics.UI.GIGtkStrut
+import qualified StatusNotifier.DBusMenu as DBusMenu
 import           StatusNotifier.Host.Service
 import qualified StatusNotifier.Item.Client as IC
 import           System.Directory
 import           System.FilePath
 import           System.Log.Logger
 import           Text.Printf
+import           Foreign.Ptr (Ptr)
 
 trayLogger :: Priority -> String -> IO ()
 trayLogger = logM "StatusNotifier.Tray"
@@ -184,7 +185,7 @@ getIconPixbufFromByteString width height byteString = do
 
 data ItemContext = ItemContext
   { contextName :: DBusTypes.BusName
-  , contextMenu :: Maybe DM.Menu
+  , contextMenuPath :: Maybe DBusTypes.ObjectPath
   , contextImage :: Gtk.Image
   , contextButton :: Gtk.EventBox
   }
@@ -234,6 +235,8 @@ buildTray Host
   trayLogger INFO "Building tray"
 
   trayBox <- Gtk.boxNew orientation 0
+  Gtk.widgetGetStyleContext trayBox >>=
+    flip Gtk.styleContextAddClass "tray-box"
   contextMap <- MV.newMVar Map.empty
 
   let getContext name = Map.lookup name <$> MV.readMVar contextMap
@@ -281,18 +284,19 @@ buildTray Host
                     info@ItemInfo { menuPath = pathForMenu
                                   , itemServiceName = serviceName
                                   , itemServicePath = servicePath
-                                  } =
+          } =
         do
-          let serviceNameStr = coerce serviceName
+          let serviceNameStr = (coerce serviceName :: String)
               servicePathStr = coerce servicePath :: String
-              serviceMenuPathStr = coerce <$> pathForMenu
               logText = printf "Adding widget for %s - %s"
                         serviceNameStr servicePathStr
 
           trayLogger INFO logText
 
-          button <- Gtk.eventBoxNew
-          Gtk.widgetAddEvents button [Gdk.EventMaskScrollMask]
+          eventBox <- Gtk.eventBoxNew
+          Gtk.widgetAddEvents eventBox [Gdk.EventMaskScrollMask]
+          Gtk.widgetGetStyleContext eventBox >>=
+            flip Gtk.styleContextAddClass "tray-icon-button"
 
           image <-
             case imageSize of
@@ -347,29 +351,34 @@ buildTray Host
           Gtk.widgetGetStyleContext image >>=
              flip Gtk.styleContextAddClass "tray-icon-image"
 
-          Gtk.containerAdd button image
-          setTooltipText button info
-
-          maybeMenu <- sequenceA $ DM.menuNew (T.pack serviceNameStr) .
-                       T.pack <$> serviceMenuPathStr
+          Gtk.containerAdd eventBox image
+          setTooltipText eventBox info
 
           let context =
                 ItemContext { contextName = serviceName
-                            , contextMenu = maybeMenu
+                            , contextMenuPath = pathForMenu
                             , contextImage = image
-                            , contextButton = button
+                            , contextButton = eventBox
                             }
-              popupItemForMenu menu = do
-                -- Cast DM.Menu to Gtk.Menu for menuPopupAtWidget
-                gtkMenu <- unsafeCastTo Gtk.Menu menu
-                Gtk.menuPopupAtWidget gtkMenu image
-                   GravitySouthWest GravityNorthWest Nothing
 
-          _ <- Gtk.onWidgetButtonPressEvent button $ \event -> do
-            button <- Gdk.getEventButtonButton event
+              popupGtkMenu gtkMenu triggerEvent = do
+                -- On Wayland (e.g. Hyprland) popups need the triggering input
+                -- event so GTK can associate the popup with the correct seat/
+                -- serial. Also, anchor to the EventBox rather than the Image
+                -- (GtkImage is typically "no-window"), or the popup may be
+                -- positioned/realized incorrectly.
+                Gtk.widgetShowAll gtkMenu
+                _ <- Gtk.onWidgetHide gtkMenu (Gtk.widgetDestroy gtkMenu)
+                evPtr <- ManagedPtr.unsafeManagedPtrCastPtr triggerEvent :: IO (Ptr Gdk.Event)
+                ManagedPtr.withTransient evPtr $ \ev ->
+                  Gtk.menuPopupAtWidget gtkMenu eventBox
+                    GravitySouthWest GravityNorthWest (Just ev)
+
+          _ <- Gtk.onWidgetButtonPressEvent eventBox $ \event -> do
+            mouseButton <- Gdk.getEventButtonButton event
             x <- round <$> Gdk.getEventButtonXRoot event
             y <- round <$> Gdk.getEventButtonYRoot event
-            action <- case button of
+            action <- case mouseButton of
               1 -> bool leftClickAction PopupMenu <$> getInfoAttr
                    itemIsMenu True serviceName
               2 -> return middleClickAction
@@ -378,9 +387,17 @@ buildTray Host
               Activate -> void $ IC.activate client serviceName servicePath x y
               SecondaryActivate -> void $ IC.secondaryActivate client
                                    serviceName servicePath x y
-              PopupMenu -> maybe (return ()) popupItemForMenu maybeMenu
+              PopupMenu -> do
+                menuPath' <- getInfoAttr menuPath Nothing serviceName
+                traverse_
+                  (\p -> catchAny
+                    (DBusMenu.buildMenu client serviceName p >>= (`popupGtkMenu` event))
+                    (\e -> trayLogger WARNING $ printf "Failed to build menu for %s: %s"
+                      (coerce serviceName :: String)
+                      (show e)))
+                  menuPath'
             return False
-          _ <- Gtk.onWidgetScrollEvent button $ \event -> do
+          _ <- Gtk.onWidgetScrollEvent eventBox $ \event -> do
             direction <- getEventScrollDirection event
             let direction' = case direction of
                                ScrollDirectionUp -> Just "vertical"
@@ -401,13 +418,13 @@ buildTray Host
 
           MV.modifyMVar_ contextMap $ return . Map.insert serviceName context
 
-          Gtk.widgetShowAll button
+          Gtk.widgetShowAll eventBox
           let packFn =
                 case alignment of
                   End -> Gtk.boxPackEnd
                   _ -> Gtk.boxPackStart
 
-          packFn trayBox button shouldExpand True 0
+          packFn trayBox eventBox shouldExpand True 0
 
       updateHandler ItemRemoved ItemInfo { itemServiceName = name }
         = getContext name >>= removeWidget
